GlowScript 2.7 VPython

#Simulador da dinâmica de muitos dipolos magnéticos interagindo entre si e com o campo magnético estático
#Essa simulação precisa ser rodada em glowscript.org
#Para rodar diretamente em glowscript, acesse: https://www.glowscript.org/#/user/jocoteles/folder/RMN/program/muitosDipolos

#from vpython import *
import random as rd

GScript = True

#{Ctes físicas:
B0 = 10.0      #campo magnetico inical [T]
B0min = 1.0    #[T]
B0max = 20.0   #[T]
Tp = 300.0      #temperatura inical [K]
Tpmin = 4.0     #[K]
Tpmax = 1000.0  #[K]
M0 = 10.0      #magnetizacao inicial [u.a.]
M0i = M0
lGammaLabels = ['1H','19F','31P','13C'] #ctes giromagnéticas [labels]
lGammaValues = [42.6, 40.1, 17.2, 10.7] #ctes giromagnéticas [MHz/T]
lAngLabels = ['45°', '90°', '135°', '180°', '225°'] #ângulos de rotação [labels]
lAngValues = [pi/4, pi/2, 3*pi/4, pi, 3*pi/2] #ângulos de rotação [rad]
N = 200                     #número de passos por volta
T = 0.5                     #período de rotação para 1H em B0 = 10 T [s]
T1 = 150*T                   #Tempo de relaxação longitudinal
T1min = 10*T
T1max = 1500*T
T2 = 100*T                   #Tempo de relaxação transversal
T2min = 10*T
T2max = 1500*T
gamma = lGammaValues[0]     #cte giromagnética [MHz/T]
rf = 2*pi/(gamma*B0*N)      #fator de reducao: 1 volta no período T para campo B0
da = gamma*B0*rf            #tamanho do passo angular para precesão de Larmor [rad]
#} Ctes físicas:


#{Flags de inicialização:
dragM = False
pausar = False
#} Flags de inicialização

vO = vector(0,0,0)
vX = vector(1,0,0)
vY = vector(0,1,0)
vZ = vector(0,0,1)
vB0i = vector(0,0,B0)
vM = vector(M0,0,0)

lEixoLabels = ['X', '-X', 'Y', '-Y'] #eixos de rotação [labels]
lEixoValues = [vX, -vX, vY, -vY] #eixos de rotação

#[ Canvas Dipolos

cDipoles = canvas(background = color.white, forward = vX, up = vZ, align='right')
cDipoles.title = '<b>Sistema de muitos spins</b>\n'

ds = 0.5    #fator de separação entre os dipolos
Nx, Ny, Nz = 3, 3, 4  #Número de dipolos em cada direção
Nxyz = Nx*Ny*Nz

#{Ctes dos dipolos:
x = 2*gamma*B0/Tp
aR = M0i/(gamma*(exp(x)-1)/(exp(x)+1))     #fator de normalização para a magnetização de equilíbrio
MEq() 
dp = 4*pi/Nxyz
m0 = C0/Nxyz
Lx, Ly, Lz = 2*(ds+1)*m0*Nx, 2*(ds+1)*m0*Ny, 2*(ds+1)*m0*Nz
#} Ctes dos dipolos

def MEq(): #cálculo da magnetização de equilíbrio e magnetização máxima
    global M0, C0
    x = 2*gamma*B0/Tp
    M0 = aR*gamma*(exp(x)-1)/(exp(x)+1)
    C0 = aR*gamma

def angleX(V):
    if mag(V) > 0:
        ph = acos(V.x/mag(V))
    else:
        ph = 0.0
    if V.y < 0:
        ph = 2*pi - ph
    return ph


def dipsLong(M):
    kl = M.z/(2*pi)
    rl = sqrt((C0/(2*pi))**2 - kl**2)
    Mt = M - vZ*M.z
    ph = angleX(Mt)
    th = 0.0
    p0 = rotate(vector(rl*cos(th),rl*sin(th),kl*th), angle=ph, axis=vZ)
    sum = vO
    for i in range(Nxyz//2):
        th += dp
        p1 = rotate(vector(rl*cos(th),rl*sin(th),kl*th), angle=ph, axis=vZ)
        dips[i].axis = p1 - p0
        sum += dips[i].axis
        p0 = p1
    return sum

def dipsTran(M):
    Mt = M - vZ*M.z
    kt = mag(Mt)/(2*pi)
    rt = sqrt((C0/(2*pi))**2 - kt**2)
    ph = angleX(Mt)
    th = 0.0
    p0 = rotate(vector(kt*th,rt*cos(th),rt*sin(th)), angle=ph, axis=vZ)
    sum = vO
    for i in range(Nxyz//2, Nxyz):
        th += dp
        p1 = rotate(vector(kt*th,rt*cos(th),rt*sin(th)), angle=ph, axis=vZ)
        dips[i].axis = p1 - p0
        sum += dips[i].axis
        p0 = p1
    return sum
    

#Inicialização dos dipolos e dos campos magnéticos:
dpos = []
for ix in range(Nx):
    qx = Lx*(ix-(Nx-1)/2)/(Nx-1)
    for iy in range(Ny):
        qy = Ly*(iy-(Ny-1)/2)/(Ny-1)
        for iz in range(Nz):
            qz = Lz*(iz-(Nz-1)/2)/(Nz-1)
            dpos.append(vector(qx,qy,qz))
rd.shuffle(dpos)

dips = []
for i in range(Nxyz):
    dips.append(arrow(canvas = cDipoles, pos = dpos[i], axis = vO, color = color.red))
    arrow(canvas = cDipoles, pos = dpos[i], axis = 2*m0*hat(vB0i), color = color.cyan, opacity = 0.5)
dsum = dipsLong(vM)
dsum += dipsTran(vM)

#] Canvas Dipolos


## [Canvas Magnetização

cMagnetization = canvas(background = color.white, forward = vX, up = vZ)
cMagnetization.title = '<b>Magnetização</b>\n'

arrow(canvas = cMagnetization, pos = -1.1*B0*vX, axis = 2.2*B0*vX, color = color.green, opacity = 0.5, shaftwidth = B0/30, headwidth = 2*B0/30, headlenght = 3*B0/30)
arrow(canvas = cMagnetization, pos = -1.1*B0*vY, axis = 2.2*B0*vY, color = color.yellow, opacity = 0.5, shaftwidth = B0/30, headwidth = 2*B0/30, headlenght = 3*B0/30)
arrow(canvas = cMagnetization, pos = -1.1*B0*vZ, axis = 2.2*B0*vZ, color = color.cyan, opacity = 0.5, shaftwidth = B0/30, headwidth = 2*B0/30, headlenght = 3*B0/30)
text(text='X', pos=1.2*B0*vX-0.75*vZ, axis=vX, up=vZ, align='center', height=1.5, depth=0.2, color=color.green)
text(text='Y', pos=1.2*B0*vY-0.75*vZ, axis=vY, up=vZ, align='center', height=1.5, depth=0.2, color=color.yellow)
text(text='Z', pos=1.2*B0*vZ-0.75*vZ, axis=vZ, up=vZ, align='center', height=1.5, depth=0.2, color=color.cyan)

aB0 = arrow(canvas = cMagnetization, pos = vO, axis = vB0i, color = color.blue, shaftwidth = B0/15, headwidth = 2*B0/15, headlenght = 3*B0/15)
aM = arrow(canvas = cMagnetization, pos = vO, axis = dsum, color = color.red, shaftwidth = M0/15, headwidth = 2*M0/15, headlenght = 3*M0/15)

def trailVec():
    return vM
atrail = attach_trail(trailVec, canvas = cMagnetization, radius=0.2, retain=N, type='points', color=color.red)
atrail.stop()

#{Movimentos do mouse
def downM():
    global dragM
    p = cMagnetization.mouse.pick
    if p != None and pausar:
        dragM = True
def moveM():
    global dragM, vM, aM
    if dragM:
        vM = mag(vM)*hat(cMagnetization.mouse.pos)
        dsum = dipsLong(vM)
        dsum += dipsTran(vM)
        aM.axis = dsum
def upM():
    global dragM
    if dragM:
        dragM = False

cMagnetization.bind('mousedown', downM)
cMagnetization.bind('mousemove', moveM)
cMagnetization.bind('mouseup', upM)
#} Movimentos do mouse

#{Funções de binding:
def bPausar(b):
    global pausar
    pausar = not pausar
    if (b.text == 'Pausar'):
        b.text = 'Executar'
    else:
        b.text = 'Pausar'

def bRotacionar(b):
    global vM
    vM = rotate(vM, angle=lAngValues[lAngLabels.index(mAng.selected)], axis=lEixoValues[lEixoLabels.index(mEix.selected)])

def bGammaCalc(g):
    global da, gamma, vM
    gamma = lGammaValues[lGammaLabels.index(g.selected)]
    da = gamma*B0*rf
    gtext.text = str(gamma) + ' MHz/T'
    MEq()
    vM = M0*vX

def bB0Calc(b0):
    global da, B0
    B0 = b0.value
    da = gamma*B0*rf
    aB0.axis = B0*hat(aB0.axis)
    tB0.text = '{:1.1f}'.format(B0*3.74/10)
    MEq()
    
def bTpCalc(T):
    global Tp
    Tp = T.value
    tTp.text = '{:1.1f}'.format(Tp)
    MEq()

def bT1(T):
    global T1, T2
    T1 = T.value
    tT1.text = '{:1.1f}'.format(T1)    
    if T1 < T2:
        T2 = T1
        sT2.value = T1

def bT2(T):
    global T1, T2
    T2 = T.value
    tT2.text = '{:1.1f}'.format(T2)    
    if T1 < T2:
        T1 = T2
        sT1.value = T2

def bMTrail(c):
    if c.checked:
        atrail.start()
    else:
        atrail.stop()
        atrail.clear()
#}} Funções de binding

#{ Widgets
button(bind=bPausar, text='Pausar', pos=cMagnetization.title_anchor)
button(bind=bRotacionar, text='Rotacionar', pos=cMagnetization.title_anchor)

cMagnetization.caption = 'Constante giromagnética (\\(\\gamma\\)) = '
gtext = wtext(text=str(lGammaValues[0])+' MHz/T  ', pos=cMagnetization.caption_anchor)
menu(bind=bGammaCalc, choices=lGammaLabels, selected=lGammaLabels[0], left=25)

cMagnetization.append_to_caption('\n\n')
sB0 = slider(bind=bB0Calc, min=B0min, max=B0max, value = B0, length=300, right=15)
cMagnetization.append_to_caption('Campo magnético (\\(B_{0}\\)): ')
tB0 = wtext(text='{:1.1f}'.format(sB0.value*3.74/10), pos=cMagnetization.caption_anchor)
cMagnetization.append_to_caption(' uT\n')

sTp = slider(bind=bTpCalc, min=Tpmin, max=Tpmax, value = Tp, length=300, right=15)
cMagnetization.append_to_caption('Temperatura (\\(T\\)): ')
tTp = wtext(text='{:1.1f}'.format(sTp.value), pos=cMagnetization.caption_anchor)
cMagnetization.append_to_caption(' K\n')

sT1 = slider(bind=bT1, min=T1min, max=T1max, value = T1, length=300, right=15)
cMagnetization.append_to_caption('Tempo de relax. longitudinal (\\(T_{1}\\)): ')
tT1 = wtext(text='{:1.1f}'.format(sT1.value), pos=cMagnetization.caption_anchor)
cMagnetization.append_to_caption(' ms\n')

sT2 = slider(bind=bT2, min=T2min, max=T2max, value = T2, length=300, right=15)
cMagnetization.append_to_caption('Tempo de relax. transversal (\\(T_{2}\\)): ')
tT2 = wtext(text='{:1.1f}'.format(sT2.value), pos=cMagnetization.caption_anchor)
cMagnetization.append_to_caption(' ms\n')

cMagnetization.append_to_caption('Ângulo de rotação (\\(\\theta\\)): ')
mAng = menu(bind=None, choices=lAngLabels, selected=lAngLabels[1], left=25)
cMagnetization.append_to_caption('   Eixo de rotação: ')
mEix = menu(bind=None, choices=lEixoLabels, selected=lEixoLabels[0], left=25)

cMagnetization.append_to_caption('\n\n')
checkbox(bind=bMTrail, text = 'Mostrar traço da magnetização', checked=False)
#} Widgets

#] Canvas Magnetização

graficoMz = graph(title='Magnetização longitudinal (Mz)', xtitle='tempo [ms]', ytitle='Amplitude [u.a.]',  fast=True, height=200)
curvaMz = gcurve(graph=graficoMz,width=2,color=color.cyan)
graficoMx = graph(title='\nMagnetização transversal (Mx)', xtitle='tempo [ms]', ytitle='Amplitude [u.a.]',  fast=True, height=200)
curvaMx = gcurve(graph=graficoMx,width=2,color=color.green)


if GScript:
    MathJax.Hub.Queue(["Typeset",MathJax.Hub])

dt = da*T/(2*pi)
t = 0
while True:
    rate(N/T)
    if ((not dragM) and (not pausar)):
        vM = rotate(vM, angle=-da, axis=aB0.axis)
        vMz = vM.z + da*(M0-vM.z)/T1
        vMT = vM-vM.z*vZ
        magT = mag(vMT)
        magT += -da*magT/T2
        vM = magT*hat(vMT) + vMz*vZ
        dsum = dipsLong(vM)
        dsum += dipsTran(vM)
        aM.axis = dsum
        t += da
        curvaMz.plot(t, dsum.z)
        curvaMx.plot(t, dsum.x)
